@@ -51,6 +51,8 @@
 #define ISO_CODES_DATADIR ISO_CODES_PREFIX "/share/xml/iso-codes"
 #define ISO_CODES_LOCALESDIR ISO_CODES_PREFIX "/share/locale"
 
+#include "default-input-sources.h"
+
 typedef struct _GnomeLocale {
         char *id;
         char *name;
@@ -113,22 +115,36 @@ normalize_codeset (const char *codeset)
         return normalized_codeset;
 }
 
-/*
- * According to http://en.wikipedia.org/wiki/Locale
- * locale names are of the form:
- * [language[_territory][.codeset][@modifier]]
+/**
+ * gnome_parse_locale:
+ * @locale: a locale string
+ * @language_codep: (out) (allow-none) (transfer full): location to
+ * store the language code, or %NULL
+ * @country_codep: (out) (allow-none) (transfer full): location to
+ * store the country code, or %NULL
+ * @codesetp: (out) (allow-none) (transfer full): location to
+ * store the codeset, or %NULL
+ * @modifierp: (out) (allow-none) (transfer full): location to
+ * store the modifier, or %NULL
+ *
+ * Extracts the various components of a locale string of the form
+ * [language[_country][.codeset][@modifier]]. See
+ * http://en.wikipedia.org/wiki/Locale.
+ *
+ * Return value: %TRUE if parsing was successful.
+ *
+ * Since: 3.8
  */
 gboolean
-gnome_parse_language_name (const char *name,
-                           char      **language_codep,
-                           char      **territory_codep,
-                           char      **codesetp,
-                           char      **modifierp)
+gnome_parse_locale (const char *locale,
+                    char      **language_codep,
+                    char      **country_codep,
+                    char      **codesetp,
+                    char      **modifierp)
 {
-        GRegex     *re;
+        static GRegex *re = NULL;
         GMatchInfo *match_info;
         gboolean    res;
-        GError     *error;
         gchar      *normalized_codeset = NULL;
         gchar      *normalized_name = NULL;
         gboolean    retval;
@@ -136,26 +152,29 @@ gnome_parse_language_name (const char *name,
         match_info = NULL;
         retval = FALSE;
 
-        error = NULL;
-        re = g_regex_new ("^(?P<language>[^_.@[:space:]]+)"
-                          "(_(?P<territory>[[:upper:]]+))?"
-                          "(\\.(?P<codeset>[-_0-9a-zA-Z]+))?"
-                          "(@(?P<modifier>[[:ascii:]]+))?$",
-                          0, 0, &error);
         if (re == NULL) {
-                g_warning ("%s", error->message);
-                goto out;
+                GError *error = NULL;
+                re = g_regex_new ("^(?P<language>[^_.@[:space:]]+)"
+                                  "(_(?P<territory>[[:upper:]]+))?"
+                                  "(\\.(?P<codeset>[-_0-9a-zA-Z]+))?"
+                                  "(@(?P<modifier>[[:ascii:]]+))?$",
+                                  0, 0, &error);
+                if (re == NULL) {
+                        g_warning ("%s", error->message);
+                        g_error_free (error);
+                        goto out;
+                }
         }
 
-        if (!g_regex_match (re, name, 0, &match_info) ||
+        if (!g_regex_match (re, locale, 0, &match_info) ||
             g_match_info_is_partial_match (match_info)) {
-                g_warning ("locale '%s' isn't valid\n", name);
+                g_warning ("locale '%s' isn't valid\n", locale);
                 goto out;
         }
 
         res = g_match_info_matches (match_info);
         if (! res) {
-                g_warning ("Unable to parse locale: %s", name);
+                g_warning ("Unable to parse locale: %s", locale);
                 goto out;
         }
 
@@ -165,13 +184,13 @@ gnome_parse_language_name (const char *name,
                 *language_codep = g_match_info_fetch_named (match_info, "language");
         }
 
-        if (territory_codep != NULL) {
-                *territory_codep = g_match_info_fetch_named (match_info, "territory");
+        if (country_codep != NULL) {
+                *country_codep = g_match_info_fetch_named (match_info, "territory");
 
-                if (*territory_codep != NULL &&
-                    *territory_codep[0] == '\0') {
-                        g_free (*territory_codep);
-                        *territory_codep = NULL;
+                if (*country_codep != NULL &&
+                    *country_codep[0] == '\0') {
+                        g_free (*country_codep);
+                        *country_codep = NULL;
                 }
         }
 
@@ -198,7 +217,7 @@ gnome_parse_language_name (const char *name,
         if (codesetp != NULL && *codesetp != NULL) {
                 normalized_codeset = normalize_codeset (*codesetp);
                 normalized_name = construct_language_name (language_codep ? *language_codep : NULL,
-                                                           territory_codep ? *territory_codep : NULL,
+                                                           country_codep ? *country_codep : NULL,
                                                            normalized_codeset,
                                                            modifierp ? *modifierp : NULL);
 
@@ -213,7 +232,6 @@ gnome_parse_language_name (const char *name,
 
  out:
         g_match_info_free (match_info);
-        g_regex_unref (re);
 
         return retval;
 }
@@ -243,23 +261,38 @@ construct_language_name (const char *language,
         return name;
 }
 
+/**
+ * gnome_normalize_locale:
+ * @locale: a locale string
+ *
+ * Gets the normalized locale string in the form
+ * [language[_country][.codeset][@modifier]] for @name.
+ *
+ * Return value: (transfer full): normalized locale string. Caller
+ * takes ownership.
+ *
+ * Since: 3.8
+ */
 char *
-gnome_normalize_language_name (const char *name)
+gnome_normalize_locale (const char *locale)
 {
         char *normalized_name;
         char *language_code;
         char *territory_code;
         char *codeset;
         char *modifier;
+        gboolean valid;
 
-        if (name[0] == '\0') {
+        if (locale[0] == '\0') {
                 return NULL;
         }
 
-        gnome_parse_language_name (name,
-                                   &language_code,
-                                   &territory_code,
-                                   &codeset, &modifier);
+        valid = gnome_parse_locale (locale,
+                                    &language_code,
+                                    &territory_code,
+                                    &codeset, &modifier);
+        if (!valid)
+                return NULL;
 
         normalized_name = construct_language_name (language_code,
                                                    territory_code,
@@ -323,15 +356,25 @@ language_name_get_codeset_details (const char  *language_name,
         g_free (old_locale);
 }
 
+/**
+ * gnome_language_has_translations:
+ * @code: an ISO 639 code string
+ *
+ * Returns %TRUE if there are translations for language @code.
+ *
+ * Return value: %TRUE if there are translations for language @code.
+ *
+ * Since: 3.8
+ */
 gboolean
-gnome_language_has_translations (const char *language_name)
+gnome_language_has_translations (const char *code)
 {
         GDir        *dir;
         char        *path;
         const char  *name;
         gboolean     has_translations;
 
-        path = g_build_filename (GNOMELOCALEDIR, language_name, "LC_MESSAGES", NULL);
+        path = g_build_filename (GNOMELOCALEDIR, code, "LC_MESSAGES", NULL);
 
         has_translations = FALSE;
         dir = g_dir_open (path, 0, NULL);
@@ -394,11 +437,11 @@ add_locale (const char *language_name,
         }
 
         locale = g_new0 (GnomeLocale, 1);
-        gnome_parse_language_name (name,
-                                   &locale->language_code,
-                                   &locale->territory_code,
-                                   &locale->codeset,
-                                   &locale->modifier);
+        gnome_parse_locale (name,
+                            &locale->language_code,
+                            &locale->territory_code,
+                            &locale->codeset,
+                            &locale->modifier);
         g_free (name);
         name = NULL;
 
@@ -1051,6 +1094,9 @@ languages_variant_init (const char *variant)
 static void
 languages_init (void)
 {
+        if (gnome_languages_map)
+                return;
+
         gnome_languages_map = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
         languages_variant_init ("iso_639");
@@ -1065,6 +1111,9 @@ territories_init (void)
         char    *buf;
         gsize    buf_len;
 
+        if (gnome_territories_map)
+                return;
+
         bindtextdomain ("iso_3166", ISO_CODES_LOCALESDIR);
         bind_textdomain_codeset ("iso_3166", "UTF-8");
 
@@ -1101,9 +1150,22 @@ territories_init (void)
         }
 }
 
+/**
+ * gnome_get_language_from_locale:
+ * @locale: a locale string
+ * @translation: (allow-none): a locale string
+ *
+ * Gets the language description for @locale. If @translation is
+ * provided the returned string is translated accordingly.
+ *
+ * Return value: (transfer full): the language description. Caller
+ * takes ownership.
+ *
+ * Since: 3.8
+ */
 char *
-gnome_get_language_from_name (const char *name,
-                              const char *locale)
+gnome_get_language_from_locale (const char *locale,
+                                const char *translation)
 {
         GString *full_language;
         char *language_code;
@@ -1114,8 +1176,8 @@ gnome_get_language_from_name (const char *name,
         char *translated_territory;
         gboolean is_utf8 = TRUE;
 
-        g_return_val_if_fail (name != NULL, NULL);
-        g_return_val_if_fail (*name != '\0', NULL);
+        g_return_val_if_fail (locale != NULL, NULL);
+        g_return_val_if_fail (*locale != '\0', NULL);
 
         translated_territory = NULL;
         translated_language = NULL;
@@ -1123,29 +1185,24 @@ gnome_get_language_from_name (const char *name,
 
         full_language = g_string_new (NULL);
 
-        if (gnome_languages_map == NULL) {
-                languages_init ();
-        }
-
-        if (gnome_territories_map == NULL) {
-                territories_init ();
-        }
+        languages_init ();
+        territories_init ();
 
         language_code = NULL;
         territory_code = NULL;
         codeset_code = NULL;
 
-        gnome_parse_language_name (name,
-                                   &language_code,
-                                   &territory_code,
-                                   &codeset_code,
-                                   NULL);
+        gnome_parse_locale (locale,
+                            &language_code,
+                            &territory_code,
+                            &codeset_code,
+                            NULL);
 
         if (language_code == NULL) {
                 goto out;
         }
 
-        translated_language = get_translated_language (language_code, locale);
+        translated_language = get_translated_language (language_code, translation);
         if (translated_language == NULL) {
                 goto out;
         }
@@ -1157,7 +1214,7 @@ gnome_get_language_from_name (const char *name,
 	}
 
         if (territory_code != NULL) {
-                translated_territory = get_translated_territory (territory_code, locale);
+                translated_territory = get_translated_territory (territory_code, translation);
         }
         if (translated_territory != NULL) {
                 g_string_append_printf (full_language,
@@ -1165,7 +1222,7 @@ gnome_get_language_from_name (const char *name,
                                         translated_territory);
         }
 
-        language_name_get_codeset_details (name, &langinfo_codeset, &is_utf8);
+        language_name_get_codeset_details (locale, &langinfo_codeset, &is_utf8);
 
         if (codeset_code == NULL && langinfo_codeset != NULL) {
                 codeset_code = g_strdup (langinfo_codeset);
@@ -1193,9 +1250,22 @@ gnome_get_language_from_name (const char *name,
         return g_string_free (full_language, FALSE);
 }
 
+/**
+ * gnome_get_country_from_locale:
+ * @locale: a locale string
+ * @translation: (allow-none): a locale string
+ *
+ * Gets the country description for @locale. If @translation is
+ * provided the returned string is translated accordingly.
+ *
+ * Return value: (transfer full): the country description. Caller
+ * takes ownership.
+ *
+ * Since: 3.8
+ */
 char *
-gnome_get_region_from_name (const char *name,
-                            const char *locale)
+gnome_get_country_from_locale (const char *locale,
+                               const char *translation)
 {
         GString *full_name;
         char *language_code;
@@ -1206,8 +1276,8 @@ gnome_get_region_from_name (const char *name,
         char *translated_territory;
         gboolean is_utf8 = TRUE;
 
-        g_return_val_if_fail (name != NULL, NULL);
-        g_return_val_if_fail (*name != '\0', NULL);
+        g_return_val_if_fail (locale != NULL, NULL);
+        g_return_val_if_fail (*locale != '\0', NULL);
 
         translated_territory = NULL;
         translated_language = NULL;
@@ -1215,29 +1285,24 @@ gnome_get_region_from_name (const char *name,
 
         full_name = g_string_new (NULL);
 
-        if (gnome_languages_map == NULL) {
-                languages_init ();
-        }
-
-        if (gnome_territories_map == NULL) {
-                territories_init ();
-        }
+        languages_init ();
+        territories_init ();
 
         language_code = NULL;
         territory_code = NULL;
         codeset_code = NULL;
 
-        gnome_parse_language_name (name,
-                                   &language_code,
-                                   &territory_code,
-                                   &codeset_code,
-                                   NULL);
+        gnome_parse_locale (locale,
+                            &language_code,
+                            &territory_code,
+                            &codeset_code,
+                            NULL);
 
         if (territory_code == NULL) {
                 goto out;
         }
 
-        translated_territory = get_translated_territory (territory_code, locale);
+        translated_territory = get_translated_territory (territory_code, translation);
         g_string_append (full_name, translated_territory);
 
 	if (is_unique_territory (territory_code)) {
@@ -1245,7 +1310,7 @@ gnome_get_region_from_name (const char *name,
 	}
 
         if (language_code != NULL) {
-                translated_language = get_translated_language (language_code, locale);
+                translated_language = get_translated_language (language_code, translation);
         }
         if (translated_language != NULL) {
                 g_string_append_printf (full_name,
@@ -1253,7 +1318,7 @@ gnome_get_region_from_name (const char *name,
                                         translated_language);
         }
 
-        language_name_get_codeset_details (name, &langinfo_codeset, &is_utf8);
+        language_name_get_codeset_details (translation, &langinfo_codeset, &is_utf8);
 
         if (codeset_code == NULL && langinfo_codeset != NULL) {
                 codeset_code = g_strdup (langinfo_codeset);
@@ -1281,8 +1346,19 @@ gnome_get_region_from_name (const char *name,
         return g_string_free (full_name, FALSE);
 }
 
+/**
+ * gnome_get_all_locales:
+ *
+ * Gets all locales.
+ *
+ * Return value: (array zero-terminated=1) (element-type utf8) (transfer full):
+ *   a newly allocated %NULL-terminated string array containing the
+ *   all locales. Free with g_strfreev().
+ *
+ * Since: 3.8
+ */
 char **
-gnome_get_all_language_names (void)
+gnome_get_all_locales (void)
 {
         GHashTableIter iter;
         gpointer key, value;
@@ -1305,3 +1381,107 @@ gnome_get_all_language_names (void)
 
         return (char **) g_ptr_array_free (array, FALSE);
 }
+
+/**
+ * gnome_get_language_from_code:
+ * @code: an ISO 639 code string
+ * @translation: (allow-none): a locale string
+ *
+ * Gets the language name for @code. If @locale is provided the
+ * returned string is translated accordingly.
+ *
+ * Return value: (transfer full): the language name. Caller takes
+ * ownership.
+ *
+ * Since: 3.8
+ */
+char *
+gnome_get_language_from_code (const char *code,
+                              const char *translation)
+{
+        g_return_val_if_fail (code != NULL, NULL);
+
+        languages_init ();
+
+        return get_translated_language (code, translation);
+}
+
+/**
+ * gnome_get_country_from_code:
+ * @code: an ISO 3166 code string
+ * @translation: (allow-none): a locale string
+ *
+ * Gets the country name for @code. If @locale is provided the
+ * returned string is translated accordingly.
+ *
+ * Return value: (transfer full): the country name. Caller takes
+ * ownership.
+ *
+ * Since: 3.8
+ */
+char *
+gnome_get_country_from_code (const char *code,
+                             const char *translation)
+{
+        g_return_val_if_fail (code != NULL, NULL);
+
+        territories_init ();
+
+        return get_translated_territory (code, translation);
+}
+
+/**
+ * gnome_get_input_source_from_locale:
+ * @locale: a locale string
+ * @type: (out) (transfer none): location to store the input source
+ * type
+ * @id: (out) (transfer none): location to store the input source
+ * identifier
+ *
+ * Gets the default input source's type and identifier for a given
+ * locale.
+ *
+ * Return value: %TRUE if a input source exists or %FALSE otherwise.
+ *
+ * Since: 3.8
+ */
+gboolean
+gnome_get_input_source_from_locale (const char  *locale,
+                                    const char **type,
+                                    const char **id)
+{
+        static GHashTable *table = NULL;
+        DefaultInputSource *dis;
+        gchar *l_code, *c_code;
+        gchar *key;
+        gint i;
+
+        g_return_val_if_fail (locale != NULL, FALSE);
+        g_return_val_if_fail (type != NULL, FALSE);
+        g_return_val_if_fail (id != NULL, FALSE);
+
+        if (!table) {
+                table = g_hash_table_new (g_str_hash, g_str_equal);
+                for (i = 0; default_input_sources[i].id; ++i) {
+                        dis = &default_input_sources[i];
+                        g_hash_table_insert (table, (gpointer) dis->locale, dis);
+                }
+        }
+
+        if (!gnome_parse_locale (locale, &l_code, &c_code, NULL, NULL))
+                return FALSE;
+
+        key = g_strconcat (l_code, "_", c_code, NULL);
+
+        dis = g_hash_table_lookup (table, key);
+        if (dis) {
+                *type = dis->type;
+                *id = dis->id;
+        }
+
+        g_free (l_code);
+        g_free (c_code);
+        g_free (key);
+
+        return dis != NULL;
+}
